import logging

from twisted.internet import ssl
from twisted.internet.protocol import Protocol, ReconnectingClientFactory

from apns.errorresponse import ErrorResponse


logger = logging.getLogger(__name__)


class GatewayClientError(Exception):
    pass


class GatewayClientNotSetError(GatewayClientError):
    pass


class GatewayClient(Protocol):
    def connectionMade(self):
        logger.debug('Gateway connection made: %s:%d', self.factory.hostname,
                     self.factory.port)
        self.factory.connectionMade(self)

    def send(self, notification):
        stream = notification.to_binary_string()
        self.transport.write(stream)

    def dataReceived(self, data):
        error = ErrorResponse()
        error.from_binary_string(data)
        self.factory.errorReceived(error)


class GatewayClientFactory(ReconnectingClientFactory):
    protocol = GatewayClient
    maxDelay = 10
    ENDPOINTS = {
        'pub': ('gateway.push.apple.com', 2195),
        'dev': ('gateway.sandbox.push.apple.com', 2195)
    }

    def __init__(self, endpoint, pem, onErrorReceived=None):
        self.hostname, self.port = self.ENDPOINTS[endpoint]
        self.client = None
        self.onErrorReceived = onErrorReceived

        with open(pem) as f:
            self.certificate = ssl.PrivateCertificate.loadPEM(f.read())

    def connectionMade(self, client):
        self.client = client

    def _onConnectionLost(self):
        self.client = None

    def errorReceived(self, error):
        logger.debug('Gateway error received: %s', error)

        if self.onErrorReceived is not None:
            self.onErrorReceived(error)

    def clientConnectionFailed(self, connector, reason):
        logger.debug('Gateway connection failed: %s',
                     reason.getErrorMessage())
        self._onConnectionLost()
        return ReconnectingClientFactory.clientConnectionFailed(self,
                                                                connector,
                                                                reason)

    def clientConnectionLost(self, connector, reason):
        logger.debug('Gateway connection lost: %s',
                     reason.getErrorMessage())
        self._onConnectionLost()
        return ReconnectingClientFactory.clientConnectionLost(self,
                                                              connector,
                                                              reason)

    @property
    def connected(self):
        return self.client is not None

    def send(self, notification):
        logger.debug('Gateway send notification')

        if self.client is None:
            raise GatewayClientNotSetError()

        self.client.send(notification)



import logging

from twisted.internet import ssl
from twisted.internet.protocol import Protocol, ReconnectingClientFactory

from apns.feedback import Feedback


logger = logging.getLogger(__name__)


class FeedbackClient(Protocol):
    def connectionMade(self):
        logger.debug('Feedback connection made: %s:%d', self.factory.hostname,
                     self.factory.port)

    def dataReceived(self, data):
        feedbacks = Feedback.from_binary_string(data)
        self.factory.feedbacksReceived(feedbacks)


class FeedbackClientFactory(ReconnectingClientFactory):
    protocol = FeedbackClient
    maxDelay = 600
    ENDPOINTS = {
        'pub': ('feedback.push.apple.com', 2196),
        'dev': ('feedback.sandbox.push.apple.com', 2196)
    }

    def __init__(self, endpoint, pem, onFeedbacksReceived=None):
        self.hostname, self.port = self.ENDPOINTS[endpoint]
        self.client = None
        self.onFeedbacksReceived = onFeedbacksReceived

        with open(pem) as f:
            self.certificate = ssl.PrivateCertificate.loadPEM(f.read())

    def feedbacksReceived(self, feedbacks):
        logger.debug('Feedbacks received: %s', feedbacks)

        if self.onFeedbacksReceived:
            self.onFeedbacksReceived(feedbacks)

    def clientConnectionFailed(self, connector, reason):
        logger.debug('Feedback connection failed: %s',
                     reason.getErrorMessage())
        return ReconnectingClientFactory.clientConnectionFailed(self,
                                                                connector,
                                                                reason)

    def clientConnectionLost(self, connector, reason):
        logger.debug('Feedback connection lost: %s',
                     reason.getErrorMessage())
        return ReconnectingClientFactory.clientConnectionLost(self,
                                                              connector,
                                                              reason)



from datetime import datetime
import binascii
import struct

from apns.utils import datetime_to_timestamp


class Feedback(object):
    FORMAT_PREFIX = '>IH'

    def __init__(self, when=None, token=None):
        self.when = when
        self.token = token

    def __str__(self):
        return '<Feedback: %s, %s>' % (self.token, self.when)


    @classmethod
    def from_binary_string(cls, stream):
        offset = 0
        length = len(stream)
        feedbacks = []

        while offset < length:
            timestamp, token_length = struct.unpack(cls.FORMAT_PREFIX,
                                                    stream[offset:offset+6])
            when = datetime.fromtimestamp(timestamp)
            offset += 6
            token = struct.unpack('>{0}s'.format(token_length),
                                  stream[offset:offset+token_length])[0]
            token = binascii.hexlify(token)
            offset += token_length
            feedbacks.append(cls(when, token))

        return feedbacks

    def to_binary_string(self):
        timestamp = datetime_to_timestamp(self.when)
        token = binascii.unhexlify(self.token)
        return struct.pack(self.FORMAT_PREFIX + '{0}s'.format(len(token)),
                           timestamp, len(token), token)





import struct

from apns.commands import ERROR_RESPONSE


class ErrorResponseError(Exception):
    pass


class ErrorResponseInvalidCommandError(ErrorResponseError):
    pass


class ErrorResponseInvalidCodeError(ErrorResponseError):
    pass


class ErrorResponse(object):
    CODES = {
        0: 'No errors encountered',
        1: 'Processing error',
        2: 'Missing device token',
        3: 'Missing topic',
        4: 'Missing payload',
        5: 'Invalid token size',
        6: 'Invalid topic size',
        7: 'Invalid payload size',
        8: 'Invalid token',
        10: 'Shutdown',
        255: 'Unknown'
    }

    FORMAT = '>BBI'
    COMMAND = ERROR_RESPONSE

    def __init__(self):
        self.code = self.name = self.identifier = None

    def __str__(self):
        return '<ErrorResponse: %s>' % self.name

    def from_binary_string(self, stream):
        command, code, identifier = struct.unpack(self.FORMAT, stream)

        if command != self.COMMAND:
            raise ErrorResponseInvalidCommandError()

        if code not in self.CODES:
            raise ErrorResponseInvalidCodeError()

        self.code = code
        self.name = self.CODES[code]
        self.identifier = identifier

    def to_binary_string(self, code, identifier):
        return struct.pack(self.FORMAT, self.COMMAND, code, identifier)



# -*- coding: utf-8 -*-
from distutils.core import setup


setup(
  name = 'twisted-apns',
  packages = ['apns'],
  version = '0.1',
  description = 'Twisted client for Apple Push Notification Service (APNs)',
  author = 'Michał Łowicki',
  author_email = 'mlowicki@opera.com',
  url = 'https://github.com/operasoftware/twisted-apns',
  download_url = 'https://github.com/operasoftware/twisted-apns/tarball/0.1',
  keywords = ['twisted', 'apns'],
  classifiers = [],
  install_requires=['Twisted', 'pyOpenSSL']
)
